# __init__.py

"""Basic lighting controller for Enttec DMX USB Pro Mk2"""

import serial
import sys
import time
import threading

start_val   = 0x7E
end_val     = 0xE7
output1     = 6
output2     = 202

class DMXConnection:
	def __init__(self, port, output = 1):
		"""Creates a connection to the DMX device.

		Parameters
		----------
		port
			On Windows, port is the port number.
        	On Linux and macOS, port is the path to the serial device.
        output: int, optional(default=1)
			DMX output connected on DMX device.
			Unless output 2 is specified, configures to output 1.

        Examples
        --------
        >>> dmx = pylightdmx.DMXConnection(4) # Windows
        Opened COM4
        >>> dmx = pylightdmx.DMXConnection('/dev/ttyUSB0') # Linux
        Opened ttyUSB0
        """
		if isinstance(port, int) == True: # Windows
			try:
				self.port = serial.Serial('COM%s' % (str(port)), 57600, timeout=1)
			except:
				print("Could not open device COM%s. Quitting application." % port)
				sys.exit(0)
		else: # Linux and macOS
			try:
				self.port = serial.Serial(port, 57600, timeout = 1)
			except:
				print("Could not open device %s. Quitting application." % port)
				sys.exit(0)
		print("Opened %s" % (self.port.portstr))
		self.dmx_frame = [0] * 512
		self.chan_list = {}
		self.output = output
	
	def set_chan(self, chan, val, auto_render = False):
		"""Sets a channel level in local channel list.

		Parameters
		----------
		chan: int
			DMX channel to be assigned a value.
			Must be between 1 and 512.
		val: int
			Value to be assigned to DMX channel.
			Must be between 0 and 255.
		auto_render: bool, optional(default=False)
			If set to true, executes the set DMX channel.

		Raises
		------
		ValueError
			If the channel is not between 1 and 512.
		"""
		if not 1 <= chan <= 512:
			raise ValueError("Invalid channel specified: %s" % str(chan))
		val = max(0, min(val, 255)) # Restrict value
		self.chan_list[chan] = val
		if auto_render == True:
			self.render()

	def render(self, clear = True, newlist = True):
		"""Executes values in channel list.

		Parameters
		----------
		clear: bool, optional(default=True)
			If set to true, clears channels not in channel list.
		newlist: bool, optional(default=True)
			If set to true, clears the channel list.
		"""
		if self.output == 2:
			label = output2
		else:
			label = output1
		if clear == True: # Clear channels not specified
			for i in range(0, 512):  
				if i not in self.chan_list.keys():
					self.dmx_frame[i] = 0
		for i in self.chan_list.keys():
			self.dmx_frame[i] = self.chan_list[i]
		packet = [
				start_val,
				label,
				len(self.dmx_frame) & 0xFF,
				(len(self.dmx_frame) >> 8) & 0xFF
				]
		packet += self.dmx_frame
		packet.append(end_val)
		self.port.write(packet)
		self.chan_list.clear()
		if newlist == True:
			self.chan_list.clear()				

	def fade(self, chan, val, secs = 3):
		"""Fades a single channel to specified value.

		Parameters
		----------
		chan: int
			DMX channel to be assigned a value.
			Must be between 1 and 512.
		val: int
			Value to be assigned to DMX channel.
			Must be between 0 and 255.
		secs: int, optional(default=3)
			Determines how many seconds to fade the channel to specified value.
		"""
		val = max(0, min(val, 255))
		orig_val = self.dmx_frame[chan]
		if val > orig_val: # Fade in
			pause = 1 / ((val - orig_val) / secs / 3)
			while True:
				value = self.dmx_frame[chan] + 3
				self.set_chan(chan, value)
				self.render(clear = False, newlist = False)
				time.sleep(pause)
				print(self.dmx_frame[chan])
				if self.dmx_frame[chan] >= val - 3:
					self.set_chan(chan, val)
					self.render(clear = False, newlist = False)
					break
		elif val < orig_val: # Fade out
			pause = 1 / (abs(val - orig_val) / secs / 3)
			while True:
				value = self.dmx_frame[chan] - 3
				self.set_chan(chan, value)
				self.render(clear = False, newlist = False)
				time.sleep(pause)
				if self.dmx_frame[chan] <= val + 3:
					self.set_chan(chan, val)
					self.render(clear = False, newlist = False)
					break
		elif val == orig_val:
			pass
		
	def generate(self, secs = 3):
		"""Fades all channels in channel list.

		Parameters
		----------
		secs: int, optional(default=3)
			Determines how many seconds to fade the channels to the specified values.
		"""
		for i in range(0, 512):  
			if i not in self.chan_list.keys():
				self.dmx_frame[i] = 0
		for chan in set(self.chan_list.keys()):
			threading.Thread(target = self.fade, args = [chan, self.chan_list[chan], secs]).start()
		main_thread = threading.current_thread()
		for t in threading.enumerate():
			if t is main_thread:
				continue
			t.join()
		self.chan_list.clear()

	def DBO(self):
		"""Sets all channels to 0, causing a dead blackout"""
		self.dmx_frame = [0] * 512
		self.render() # Auto renders
		
	def close(self):
		"""Closes connection to DMX device."""
		self.port.close()


# fixtures.py

"""Adds fixture definitions for use with pylightdmx"""

import json
import os
from PIL import ImageColor
import pylightdmx

available_controls = ["intensity", "pan", "tilt", "speed", "macros", "focus", "strobe"]
rgb_channels = ["red", "green", "blue"]

class Fixture():
	def __init__(self, connection, brand, model, address):
		"""Opens JSON file containing fixture definition.

		Parameters
		----------
		connection: obj
			The DMX connection opened by pylightdmx for the DMX device in use.
        brand: str
			The brand of the fixture.
			Must correspond to the directory the JSON file is located in.
		model: str
			The name of the fixture.
			Must correspond to the file name of the JSON file.
		address: int
			The first DMX channel to be used by the fixture.

        Examples
        --------
        >>> LED = fixtures.Fixture(dmx, "Generic", "RGB", 1)
        """
		path = os.path.join(os.path.dirname(__file__), "rigs", "fixtures", brand, model + ".json")
		with open(path, 'r') as f:
			self.data = json.load(f)
		self.address = address
		self.link = connection
		self.name = self.data["shortName"]
		self.speed_offset = {}
		self.macro_offset = {}

	def __str__(self):
		return "{}, Address = {}".format(self.name, self.address)

	def rgb_control(self):
		"""Initialises the ability to use the RGB channels of the fixture."""
		self.r_offset = self.data["availableChannels"]["red"]["offset"]
		self.g_offset = self.data["availableChannels"]["green"]["offset"]
		self.b_offset = self.data["availableChannels"]["blue"]["offset"]

	def set_rgb(self, r, g, b):
		"""Sets the colour of the fixture using RGB values.

		Parameters
		----------
		r: int
			Value to set the red value on the fixture to.
		g: int
			Value to set the green value on the fixture to.
		b: int
			Value to set the blue value on the fixture to.

		Examples
		--------
		>>> LED = fixtures.Fixture(dmx, "Generic", "RGB", 1)
		>>> LED.rgb_control()
		>>> LED.set_rgb(0, 128, 0)
		>>> dmx.render()
		"""
		self.link.set_chan(self.address + self.r_offset, r)
		self.link.set_chan(self.address + self.g_offset, g)
		self.link.set_chan(self.address + self.b_offset, b)

	def set_colour(self, colour):
		"""Sets the colour of the fixture using the name of a colour.

		Parameters
		----------
		colour: str
			Colour to set the fixture to.
		"""
		r, g, b = ImageColor.getcolor(colour, "RGB")
		self.set_rgb(r, g, b)

	def intensity(self):
		"""Initialises the ability to use the intensity channel of the fixture."""
		self.intensity_offset = self.data["availableChannels"]["intensity"]["offset"]

	def set_intensity(self, val):
		"""Sets the intensity of the fixture.

		Parameters
		----------
		val
			Value(int) or percentage(str) to set the intensity of the fixture to.

		Examples
		--------
		>>> Par = fixtures.Fixture(dmx, "Generic", "Dimmer", 1)
		>>> Par.intensity()
		>>> Par.set_intensity(255)
		>>> dmx.render()
		>>> Par.set_intensity("75%")
		>>> dmx.render()
		"""
		if isinstance(val, int) == True:
			self.link.set_chan(self.address + self.intensity_offset, val)
		elif "%" in val:
			percent = float(val.strip('%'))/100
			val = int(percent * 255)
			self.link.set_chan(self.address + self.intensity_offset, val)

	def strobe(self):
		"""Initialises the ability to use the strobe channel of the fixture."""
		self.strobe_offset = self.data["availableChannels"]["strobe"]["offset"]

	def set_strobe(self, val):
		"""Sets the strobe of the fixture.

		Parameters
		----------
		val
			Value(int) or percentage(str) to set the strobe of the fixture to.
		"""
		if isinstance(val, int) == True:
			self.link.set_chan(self.address + self.strobe_offset, val)
		elif "%" in val:
			percent = float(val.strip('%'))/100
			val = int(percent * 255)
			self.link.set_chan(self.address + self.strobe_offset, val)

	def focus(self):
		"""Initialises the ability to use the focus channel of the fixture."""
		self.focus_offset = self.data["availableChannels"]["focus"]["offset"]

	def set_focus(self, val):
		"""Sets the focus of the fixture.

		Parameters
		----------
		val
			Value(int) or percentage(str) to set the focus of the fixture to.
		"""
		if isinstance(val, int) == True:
			self.link.set_chan(self.address + self.focus_offset, val)
		elif "%" in val:
			percent = float(val.strip('%'))/100
			val = int(percent * 255)
			self.link.set_chan(self.address + self.focus_offset, val)

	def pan(self):
		"""Initialises the ability to use the pan channel of the fixture."""
		self.pan_offset = self.data["availableChannels"]["pan"]["offset"]
		self.range = self.data["availableChannels"]["pan"]["range"]
		self.dmx_per_deg = 255/self.range

	def set_pan(self, val):
		"""Sets the pan of the fixture.

		Parameters
		----------
		val
			Value(int) or angle(str) to set the pan of the fixture to.

		Examples
		--------
		>>> MH = fixtures.Fixture(dmx, "iSolution", "iMove-8s", 1)
		>>> MH.pan()
		>>> MH.set_pan(255)
		>>> dmx.render()
		>>> MH.set_pan("180*")
		>>> dmx.render()
		"""
		if isinstance(val, int) == True:
			 self.link.set_chan(self.address + self.pan_offset, val)
		elif "*" in val:
			angle = float(val.strip("*"))
			val = int(angle * self.dmx_per_deg)
			self.link.set_chan(self.address + self.pan_offset, val) 

	def tilt(self):
		"""Initialises the ability to use the tilt channel of the fixture."""
		self.tilt_offset = self.data["availableChannels"]["tilt"]["offset"]
		self.range = self.data["availableChannels"]["tilt"]["range"]
		self.dmx_per_deg = 255/self.range

	def set_tilt(self, val):
		"""Sets the tilt of the fixture.

		Parameters
		----------
		val
			Value(int) or angle(str) to set the tilt of the fixture to.
		"""
		if isinstance(val, int) == True:
			 self.link.set_chan(self.address + self.tilt_offset, val)
		elif "*" in val:
			angle = float(val.strip("*"))
			val = int(angle * self.dmx_per_deg)
			self.link.set_chan(self.address + self.tilt_offset, val)

	def speed(self, name):
		"""Initialises the ability to use a speed function of the fixture.

		Parameters
		----------
		name: str
			Name of speed function to initialise.

		Examples
		--------
		>>> MH = fixtures.Fixture(dmx, "Rave", "Mini_Spot_Moving_Head", 1)
		>>> MH.speed("Pan/Tilt Speed")
		"""
		self.speed_offset[name] = self.data["availableChannels"][name]["offset"]

	def set_speed(self, name, val):
		"""Sets the value of a speed function of the fixture.

		Parameters
		----------
		name: str
			Name of the speed function to assign a value to.
		val: int
			Value to set the speed function to.

		Examples
		--------
		>>> MH = fixtures.Fixture(dmx, "Rave", "Mini_Spot_Moving_Head", 1)
		>>> MH.speed("Pan/Tilt Speed")
		>>> MH.set_speed("Pan/Tilt Speed", 255)
		>>> dmx.render()
		"""
		self.link.set_chan(self.address + self.speed_offset[name], val)

	def macros(self, name):
		"""Initialises the ability to use a macro channel of the fixture.

		Parameters
		----------
		name: str
			Name of macro function to initialise.

		Examples
		--------
		>>> MH = fixtures.Fixture(dmx, "Rave", "Mini_Spot_Moving_Head", 1)
		>>> MH.macros("Colour Macros")
		"""
		self.macro_offset[name] = self.data["availableChannels"][name]["offset"]

	def set_macro(self, name, label):
		"""Sets a macro of the fixture.

		Parameters
		----------
		name: str
			Name of the macro to assign a value to.
		label: str
			Value to set the macro to. 

		Examples
		--------
		>>> MH = fixtures.Fixture(dmx, "Rave", "Mini_Spot_Moving_Head", 1)
		>>> MH.macros("Colour Macros")
		>>> MH.set_macro("Colour Macros", "magenta")
		>>> dmx.render()
		"""
		val = self.data["availableChannels"][name]["capabilities"][label]["startVal"] 
		self.link.set_chan(self.address + self.macro_offset[name], val)

	def config(self):
		"""Initialises all available channels of the fixture."""
		for i in available_controls:
			for d in self.data["availableChannels"]:
				category = self.data["availableChannels"][d]["type"]
				if  category == i:
					if category == "macros" or category == "speed":
						getattr(self, i)(str(d))
					else:
						getattr(self, i)()
		if all(x in self.data["availableChannels"] for x in rgb_channels):
			self.rgb_control()
		
	def list_channels(self):
		"""Lists available channels of the fixture."""
		return list(self.data["availableChannels"].keys())


from setuptools import setup, find_packages

setup(name = "pylightdmx",
	version = "1.0.2",
	description = "Enttec DMX USB Pro Mk2 Lighting Controller",
	packages = find_packages(),
	license = "GPLV3",
	author = "ayork02",
	author_email = "ayork02@outlook.com",
	url = "https://github.com/ayork02/pylightdmx",
	classifiers = [
		"Development Status :: 3 - Alpha",
		"License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
		"Programming Language :: Python :: 3"
	],
	install_requires = ["pyserial", "Pillow"]
	)


import json
import os
import pylightdmx
from pylightdmx import fixtures

class FixtureGroup:
	def __init__(self, connection, rig, name):
		"""Inititialises fixture group from rig definition.
		
		Parameters
		----------
		connection: obj
			The DMX connection opened by pylightdmx for the DMX device in use.
        rig: str
			The name of the rig.
			Must correspond to the file name of the JSON file.
		name: str
			The name of the fixture group.
			Must correspond to the fixture group name in the rig defintion.
		
        Examples
        --------
        >>> g = rigs.FixtureGroup(dmx, "example_rig", "Dimmers")
        """
		self.link = connection
		self.g = {}
		for fixture in rig.rig_data["groups"][name]:
			self.g[fixture] = fixtures.Fixture(connection, rig.rig_data["fixtures"][fixture]["brand"], rig.rig_data["fixtures"][fixture]["model"], rig.rig_data["fixtures"][fixture]["address"])
			self.g[fixture].config()
	
	def rgb_control(self):
		"""Initialises the ability to use the RGB channels of the fixture group."""
		for fixture in self.g.keys():
			self.g[fixture].rgb_control()
		
	def set_rgb(self, r, g, b):
		"""Sets the colour of the fixture group using RGB values.

		Parameters
		----------
		r: int
			Value to set the red value on the fixture group to.
		g: int
			Value to set the green value on the fixture group to.
		b: int
			Value to set the blue value on the fixture group to.

		Examples
		--------
		>>> LEDs = rigs.FixtureGroup(dmx, "example_rig", "LEDs")
		>>> LEDs.rgb_control()
		>>> LEDs.set_rgb(0, 128, 0)
		>>> dmx.render()
		"""
		for fixture in self.g.keys():
			self.g[fixture].set_rgb(r, g, b)
	
	def set_colour(self, colour):
		"""Sets the colour of the fixture group using the name of a colour.

		Parameters
		----------
		colour: str
			Colour to set the fixture group to.
		"""
		for fixture in self.g.keys():
			self.g[fixture].set_colour(colour)
		
	def intensity(self):
		"""Initialises the ability to use the intensity channel of the fixture group."""
		for fixture in self.g.keys():
			self.g[fixture].intensity()
		
	def set_intensity(self, val):
		"""Sets the intensity of the fixture group.

		Parameters
		----------
		val
			Value(int) or percentage(str) to set the intensity of the fixture group to.
		"""
		for fixture in self.g.keys():
			self.g[fixture].set_intensity(val)

	def strobe(self):
		"""Initialises the ability to use the strobe channel of the fixture group."""
		for fixture in self.g.keys():
			self.g[fixture].strobe()
		
	def set_strobe(self, val):
		"""Sets the strobe of the fixture group.

		Parameters
		----------
		val
			Value(int) or percentage(str) to set the strobe of the fixture group to.
		"""
		for fixture in self.g.keys():
			self.g[fixture].strobe(val)

	def focus(self):
		"""Initialises the ability to use the focus channel of the fixture group."""
		for fixture in self.g.keys():
			self.g[fixture].focus()
		
	def set_focus(self, val):
		"""Sets the focus of the fixture group.

		Parameters
		----------
		val
			Value(int) or percentage(str) to set the focus of the fixture group to.
		"""
		for fixture in self.g.keys():
			self.g[fixture].set_focus(val)
		
	def pan(self):
		"""Initialises the ability to use the pan channel of the fixture group."""
		for fixture in self.g.keys():
			self.g[fixture].pan()

	def set_pan(self, val):
		"""Sets the pan of the fixture group.

		Parameters
		----------
		val
			Value(int) or angle(str) to set the pan of the fixture group to.
		"""
		for fixture in self.g.keys():
			self.g[fixture].set_pan(val)
		
	def tilt(self):
		"""Initialises the ability to use the tilt channel of the fixture group."""
		for fixture in self.g.keys():
			self.g[fixture].tilt()
		
	def set_tilt(self, val):
		"""Sets the tilt of the fixture group.

		Parameters
		----------
		val
			Value(int) or angle(str) to set the tilt of the fixture group to.
		"""
		for fixture in self.g.keys():
			self.g[fixture].set_tilt(val)

	def speed(self, name):
		"""Initialises the ability to use a speed function of the fixture group.

		Parameters
		----------
		name: str
			Name of speed function to initialise.
		"""
		for fixture in self.g.keys():
			self.g[fixture].speed(name)

	def set_speed(self, name, val):
		"""Sets the value of a speed function of the fixture group.

		Parameters
		----------
		name: str
			Name of the speed function to assign a value to.
		val: int
			Value to set the speed function to.
		"""
		for fixture in self.g.keys():
			self.g[fixture].set_speed(name, val)

	def macros(self, name):
		"""Initialises the ability to use a macro channel of the fixture group.

		Parameters
		----------
		name: str
			Name of macro function to initialise.
		"""
		for fixture in self.g.keys():
			self.g[fixture].macros(name)

	def set_macro(self, name, label):
		"""Sets a macro of the fixture group.

		Parameters
		----------
		name: str
			Name of the macro to assign a value to.
		label: str
			Value to set the macro to. 
		"""
		for fixture in self.g.keys():
			self.g[fixture].set_macro(name, label)	

	def config(self):
		"""Initialises all available channels of the fixture group."""
		for fixture in self.g.keys():
			self.g[fixture].config()

		
class Rig:
	def __init__(self, connection, name):
		"""Opens JSON file containing rig definition and initialises all fixtures and fixture groups in the rig.

		Parameters
		----------
		connection: obj
			The DMX connection opened by pylightdmx for the DMX device in use.
        name: str
			The name of the rig.
			Must correspond to the file name of the JSON file.
		
        Examples
        --------
        >>> r = rigs.Rig(dmx, "example_rig")
        >>> r.f["LED1"].set_rgb(255, 0, 0) # Use of a fixture in rig
        >>> r.g["Dimmers"].set_intensity(255) # Use of a group in the rig
        """
		path = os.path.join(os.path.dirname(__file__), "rigs", name + ".json")
		with open(path, 'r') as f:
			self.rig_data = json.load(f)
		self.f = {}
		self.g = {}
		for name, d in self.rig_data["fixtures"].items():
			self.f[name] = fixtures.Fixture(connection, d["brand"], d["model"], d["address"])
			self.f[name].config()
		for name in self.rig_data["groups"]:
			self.g[name] = FixtureGroup(connection, self, name)
			self.g[name].config()